#ifdef USE_GRAPHIC
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "config.h"
#include "const.h"
#include "image.h"

static struct image images[GRAPHIC_IMAGE_MAXNUM];

int image_is_draw_line(image_t image, image_draw_count_t count, int y)
{
  return (image->draw.line[y] >= count) ? 1 : 0;
}

int image_is_draw_column(image_t image, image_draw_count_t count, int x, int y)
{
  int i;
  i = image->draw.column_num * y + (x / DRAW_COLUMN_PER);
  return (image->draw.column[i] >= count) ? 1 : 0;
}

int image_is_draw_pixel(image_t image, image_draw_count_t count, int x, int y)
{
  return (image->draw.pixel[image->width * y + x] >= count) ? 1 : 0;
}

int image_set_draw_line(image_t image, image_draw_count_t count, int y)
{
  image->draw.line[y] = count;
  return 0;
}

int image_set_draw_column(image_t image, image_draw_count_t count, int x, int y)
{
  int i;
  i = image->draw.column_num * y + (x / DRAW_COLUMN_PER);
  image->draw.column[i] = count;
  return 0;
}

int image_set_draw_pixel(image_t image, image_draw_count_t count, int x, int y)
{
  image->draw.pixel[image->width * y + x] = count;
  return 0;
}

int image_get_color(image_t image, int x, int y)
{
  if ((x < 0) || (x > image->width  - 1) ||
      (y < 0) || (y > image->height - 1))
    return -1;
  return image->bitmap[image->width * y + x];
}

int image_set_color(image_t image, int x, int y, int color, unsigned int flags)
{
  if ((x < 0) || (x > image->width  - 1) ||
      (y < 0) || (y > image->height - 1))
    return -1;

  switch (flags & NLL_G_MODE_MASK) {
  case NLL_G_AND: color &= image_get_color(image, x, y); break;
  case NLL_G_OR:  color |= image_get_color(image, x, y); break;
  case NLL_G_XOR: color ^= image_get_color(image, x, y); break;
  case NLL_G_SET:
  default:
    break;
  }

  image->bitmap[image->width * y + x] = color;

  return 0;
}

int image_get_index(image_t image)
{
  if (!image)
    return -1;
  return image - images;
}

image_t image_get_image(int index)
{
  image_t image;

  if (index > GRAPHIC_IMAGE_MAXNUM)
    return NULL;

  if (index < 0) {
    for (index = 0; index < GRAPHIC_IMAGE_MAXNUM; index++) {
      image = &images[index];
      if (image->status == IMAGE_STATUS_FREE)
	image_done(image);
      if (image->status == IMAGE_STATUS_NONE)
	break;
    }
    if (index == GRAPHIC_IMAGE_MAXNUM)
      return NULL;
  }

  return &images[index];
}

int image_reset_draw(image_t image)
{
  memset(image->draw.line, 0, image->height * sizeof(image_draw_count_t));
  memset(image->draw.column, 0, image->draw.column_num * image->height * sizeof(image_draw_count_t));
  memset(image->draw.pixel, 0, image->width * image->height * sizeof(image_draw_count_t));
  return 0;
}

int image_done(image_t image)
{
  image->status = IMAGE_STATUS_NONE;

  if (image->draw.pixel)  free(image->draw.pixel);
  if (image->draw.column) free(image->draw.column);
  if (image->draw.line)   free(image->draw.line);
  if (image->bitmap)      free(image->bitmap);
  memset(image, 0, sizeof(*image));

  return 0;
}

int image_init(image_t image, int width, int height, int *bitmap)
{
  memset(image, 0, sizeof(*image));

  image->width  = width;
  image->height = height;

  image->size = image->width * image->height * sizeof(*(image->bitmap));
  if (bitmap) {
    image->bitmap = bitmap;
  } else {
    image->bitmap = malloc(image->size);
    if (!image->bitmap)
      goto err;
    memset(image->bitmap, 0, image->size);
  }

  image->draw.column_num = (image->width + DRAW_COLUMN_PER - 1) / DRAW_COLUMN_PER;

  image->draw.line   = malloc(image->height * sizeof(image_draw_count_t));
  image->draw.column = malloc(image->draw.column_num * image->height * sizeof(image_draw_count_t));
  image->draw.pixel  = malloc(image->width * image->height * sizeof(image_draw_count_t));
  if (!image->draw.line || !image->draw.column || !image->draw.pixel)
    goto err;

  image_reset_draw(image);

  image->draw.count = 0;

  image->status = IMAGE_STATUS_ACTIVE;

  return 0;

err:
  image_done(image);
  return -1;
}

image_t image_get(int width, int height, int *bitmap)
{
  image_t image;

  image = image_get_image(-1);
  if (!image)
    return NULL;

  if (image_init(image, width, height, bitmap) < 0)
    return NULL;

  return image;
}
#endif
